#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from os.path import abspath
from datetime import datetime
from uuid import uuid4
import logging

from django.contrib.gis.geos import GEOSGeometry

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import config
from eoxserver.core.util.rect import Rect
from eoxserver.contrib import gdal, osr
from eoxserver.contrib.vrt import VRTBuilder
from eoxserver.services.ows.version import Version
from eoxserver.services.result import ResultFile, ResultBuffer
from eoxserver.services.ows.wcs.v20.encoders import WCS20EOXMLEncoder
from eoxserver.services.exceptions import (
    RenderException, OperationNotSupportedException
)
from eoxserver.processing.gdal import reftools


logger = logging.getLogger(__name__)


class GDALReferenceableDatasetRenderer(object):

    versions = (Version(2, 0),)

    def supports(self, params):
        return (
            params.coverage.grid.is_referenceable and params.version in self.versions
        )

    def render(self, params):
        # get the requested coverage, data items and range type.
        coverage = params.coverage
        data_locations = coverage.arraydata_locations
        range_type = coverage.range_type

        subsets = params.subsets

        # GDAL source dataset. Either a single file dataset or a composed VRT
        # dataset.
        src_ds = self.get_source_dataset(
            coverage, data_locations, range_type
        )

        # retrieve area of interest of the source image according to given
        # subsets
        src_rect, dst_rect = self.get_source_and_dest_rect(src_ds, subsets)

        # deduct "native" format of the source image
        native_format = data_locations[0].format if len(data_locations) == 1 else None

        # get the requested image format, which defaults to the native format
        # if available
        frmt = params.format or native_format

        if not frmt:
            raise RenderException("No format specified.", "format")

        if params.scalefactor is not None or params.scales:
            raise RenderException(
                "ReferenceableDataset cannot be scaled.",
                "scalefactor" if params.scalefactor is not None else "scale"
            )

        maxsize = WCSConfigReader(get_eoxserver_config()).maxsize
        if maxsize is not None:
            if maxsize < dst_rect.size_x or maxsize < dst_rect.size_y:
                raise RenderException(
                    "Requested image size %dpx x %dpx exceeds the allowed "
                    "limit maxsize=%dpx." % (
                        dst_rect.size_x, dst_rect.size_y, maxsize
                    ), "size"
                )

        # perform subsetting either with or without rangesubsetting
        subsetted_ds = self.perform_subset(
            src_ds, range_type, src_rect, dst_rect, params.rangesubset
        )

        # encode the processed dataset and save it to the filesystem
        out_ds, out_driver = self.encode(
            subsetted_ds, frmt, getattr(params, "encoding_params", {})
        )

        driver_metadata = out_driver.GetMetadata_Dict()
        mime_type = driver_metadata.get("DMD_MIMETYPE")
        extension = driver_metadata.get("DMD_EXTENSION")

        time_stamp = datetime.now().strftime("%Y%m%d%H%M%S")
        filename_base = "%s_%s" % (coverage.identifier, time_stamp)

        result_set = [
            ResultFile(
                path, mime_type, "%s.%s" % (filename_base, extension),
                ("cid:coverage/%s" % coverage.identifier) if i == 0 else None
            ) for i, path in enumerate(out_ds.GetFileList())
        ]

        if params.mediatype and params.mediatype.startswith("multipart"):
            reference = "cid:coverage/%s" % result_set[0].filename

            if subsets.has_x and subsets.has_y:
                footprint = GEOSGeometry(reftools.get_footprint_wkt(out_ds))
                if not subsets.srid:
                    extent = footprint.extent
                else:
                    extent = subsets.xy_bbox
                encoder_subset = (
                    subsets.srid, src_rect.size, extent, footprint
                )
            else:
                encoder_subset = None

            encoder = WCS20EOXMLEncoder()
            content = encoder.serialize(
                encoder.encode_referenceable_dataset(
                    coverage, range_type, reference, mime_type, encoder_subset
                )
            )
            result_set.insert(0, ResultBuffer(content, encoder.content_type))

        return result_set

    def get_source_dataset(self, coverage, data_items, range_type):
        if len(data_items) == 1:
            return gdal.OpenShared(data_items[0].path)
        else:
            vrt = VRTBuilder(
                coverage.size_x, coverage.size_y,
                vrt_filename=temp_vsimem_filename()
            )

            # sort in ascending order according to semantic
            data_items = sorted(data_items, key=(lambda d: d.semantic))

            compound_index = 0
            for data_item in data_items:
                path = data_item.path

                # iterate over all bands of the data item
                indices = self._data_item_band_indices(data_item)
                for set_index, item_index in indices:
                    if set_index != compound_index + 1:
                        raise ValueError
                    compound_index = set_index

                    band = range_type[set_index]
                    vrt.add_band(band.data_type)
                    vrt.add_simple_source(
                        set_index, path, item_index
                    )

            return vrt.dataset

    def get_source_and_dest_rect(self, dataset, subsets):
        size_x, size_y = dataset.RasterXSize, dataset.RasterYSize
        image_rect = Rect(0, 0, size_x, size_y)

        if not subsets:
            subset_rect = image_rect

        # pixel subset
        elif subsets.srid is None:  # means "imageCRS"
            minx, miny, maxx, maxy = subsets.xy_bbox

            minx = int(minx) if minx is not None else image_rect.offset_x
            miny = int(miny) if miny is not None else image_rect.offset_y
            maxx = int(maxx) if maxx is not None else image_rect.upper_x
            maxy = int(maxy) if maxy is not None else image_rect.upper_y

            subset_rect = Rect(minx, miny, maxx-minx+1, maxy-miny+1)

        # subset in geographical coordinates
        else:
            vrt = VRTBuilder(*image_rect.size)
            vrt.copy_gcps(dataset)

            options = reftools.suggest_transformer(dataset)

            subset_rect = reftools.rect_from_subset(
                vrt.dataset, subsets.srid, *subsets.xy_bbox, **options
            )

        # check whether or not the subsets intersect with the image
        if not image_rect.intersects(subset_rect):
            raise RenderException("Subset outside coverage extent.", "subset")

        src_rect = subset_rect  # & image_rect # TODO: why no intersection??
        dst_rect = src_rect - subset_rect.offset

        return src_rect, dst_rect

    def perform_subset(self, src_ds, range_type, subset_rect, dst_rect,
                       rangesubset=None):
        vrt = VRTBuilder(*subset_rect.size)

        input_bands = list(range_type)

        # list of band indices/names. defaults to all bands
        if rangesubset:
            subset_bands = rangesubset.get_band_indices(range_type, 1)
        else:
            subset_bands = range(1, len(range_type) + 1)

        for dst_index, src_index in enumerate(subset_bands, start=1):
            input_band = input_bands[src_index-1]
            vrt.add_band(input_band.data_type)
            vrt.add_simple_source(
                dst_index, src_ds, src_index, subset_rect, dst_rect
            )

        vrt.copy_metadata(src_ds)
        vrt.copy_gcps(src_ds, subset_rect)

        return vrt.dataset

    def encode(self, dataset, frmt, encoding_params):
        options = ()
        if frmt == "image/tiff":
            options = _get_gtiff_options(**encoding_params)

        args = [
            ("%s=%s" % key, value) for key, value in options
        ]

        path = "/tmp/%s" % uuid4().hex
        out_driver = gdal.GetDriverByName("GTiff")
        return out_driver.CreateCopy(path, dataset, True, args), out_driver


def index_of(iterable, predicate, default=None, start=1):
    for i, item in enumerate(iterable, start):
        if predicate(item):
            return i
    return default


def temp_vsimem_filename():
    return "/vsimem/%s" % uuid4().hex


def _get_gtiff_options(compression=None, jpeg_quality=None,
                       predictor=None, interleave=None, tiling=False,
                       tilewidth=None, tileheight=None):

    logger.info("Applying GeoTIFF parameters.")

    if compression:
        if compression.lower() == "huffman":
            compression = "CCITTRLE"
        yield ("COMPRESS", compression.upper())

    if jpeg_quality is not None:
        yield ("JPEG_QUALITY", str(jpeg_quality))

    if predictor:
        pr = ["NONE", "HORIZONTAL", "FLOATINGPOINT"].index(predictor.upper())
        if pr == -1:
            raise ValueError("Invalid compression predictor '%s'." % predictor)
        yield ("PREDICTOR", str(pr + 1))

    if interleave:
        yield ("INTERLEAVE", interleave)

    if tiling:
        yield ("TILED", "YES")
        if tilewidth is not None:
            yield ("BLOCKXSIZE", str(tilewidth))
        if tileheight is not None:
            yield ("BLOCKYSIZE", str(tileheight))


class WCSConfigReader(config.Reader):
    section = "services.ows.wcs"
    maxsize = config.Option(type=int, default=None)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

""" This module is an extension of the :mod:`iteratortools` module and provides
additional helpers.
"""

try:
    from itertools import izip
except ImportError:
    izip = zip

from itertools import tee


def pairwise(iterable):
    "s -> (s0,s1), (s2,s3), (s4, s5), ..."
    a = iter(iterable)
    return izip(a, a)


def pairwise_iterative(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return izip(a, b)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage product types. This command uses sub-commands for the
        specific tasks: create, delete
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create')
        delete_parser = self.add_subparser(parser, 'delete')
        list_parser = self.add_subparser(parser, 'list')

        for parser in [create_parser, delete_parser]:
            parser.add_argument(
                'name', nargs=1, help='The product type name. Mandatory.'
            )

        create_parser.add_argument(
            '--coverage-type', '-c',
            action='append', dest='coverage_type_names', default=[],
            help=(
            )
        )
        create_parser.add_argument(
            '--mask-type', '-m',
            action='append', dest='mask_type_names', default=[],
            help=(
            )
        )
        create_parser.add_argument(
            '--validity-mask-type',
            action='append', dest='validity_mask_type_names', default=[],
            help=(
            )
        )
        create_parser.add_argument(
            '--browse-type', '-b',
            action='append', dest='browse_type_names', default=[],
            help=(
            )
        )

        delete_parser.add_argument(
            '--force', '-f', action='store_true', default=False,
            help='Also remove all products associated with that type.'
        )

        list_parser.add_argument(
            '--no-detail', action="store_false", default=True, dest='detail',
            help="Disable the printing of details of the product type."
        )

    @transaction.atomic
    def handle(self, subcommand, *args, **kwargs):
        """ Dispatch sub-commands: create, delete, list.
        """
        if subcommand == "create":
            self.handle_create(kwargs.pop('name')[0], *args, **kwargs)
        elif subcommand == "delete":
            self.handle_delete(kwargs.pop('name')[0], *args, **kwargs)
        elif subcommand == "list":
            self.handle_list(*args, **kwargs)

    def handle_create(self, name, coverage_type_names, mask_type_names,
                      validity_mask_type_names, browse_type_names,
                      *args, **kwargs):
        """ Handle the creation of a new product type.
        """

        product_type = models.ProductType.objects.create(name=name)

        for coverage_type_name in coverage_type_names:
            try:
                coverage_type = models.CoverageType.objects.get(
                    name=coverage_type_name
                )
                product_type.allowed_coverage_types.add(coverage_type)
            except models.CoverageType.DoesNotExist:
                raise CommandError(
                    'Coverage type %r does not exist' % coverage_type_name
                )

        for mask_type_name in mask_type_names:
            models.MaskType.objects.create(
                name=mask_type_name, product_type=product_type
            )

        for mask_type_name in validity_mask_type_names:
            models.MaskType.objects.create(
                name=mask_type_name, product_type=product_type,
                validity=True
            )

        for browse_type_name in browse_type_names:
            models.BrowseType.objects.create(
                name=browse_type_name, product_type=product_type
            )

        print('Successfully created product type %r' % name)

    def handle_delete(self, name, force, **kwargs):
        """ Handle the deletion of a product type
        """

        try:
            product_type = models.ProductType.objects.get(name=name)
        except models.ProductType.DoesNotExist:
            raise CommandError('No such product type %r' % name)

        if force:
            products = models.Product.objects.filter(product_type=product_type)
            for product in products:
                product.delete()

        product_type.delete()
        # TODO force
        print('Successfully deleted product type %r' % name)

    def handle_list(self, detail, *args, **kwargs):
        """ Handle the listing of product types
        """
        for product_type in models.ProductType.objects.all():
            print(product_type.name)
            if detail:
                for coverage_type in product_type.allowed_coverage_types.all():
                    print("\t%s" % coverage_type.name)



# -----------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# -----------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

"""\
This module contains a set of handler base classes which shall help to
implement a specific handler. Interface methods need to be overridden in order
to work, default methods can be overidden.
"""
from django.db.models import Q

from eoxserver.resources.coverages import models
from eoxserver.services.result import to_http_response
from eoxserver.services.ows.wcs.parameters import WCSCapabilitiesRenderParams
from eoxserver.services.exceptions import (
    NoSuchCoverageException, OperationNotSupportedException
)
from eoxserver.services.ows.wcs.renderers import (
    get_capabilities_renderer, get_coverage_description_renderer,
    get_coverage_renderer,
)
from eoxserver.render.coverage.objects import Coverage, Mosaic
from eoxserver.services.ecql import parse, to_filter
from eoxserver.services import filters


class WCSGetCapabilitiesHandlerBase(object):
    """ Base for Coverage description handlers.
    """

    service = "WCS"
    request = "GetCapabilities"

    index = 0

    def get_decoder(self, request):
        """ Interface method to get the correct decoder for this request.
        """

    def lookup_coverages(self, decoder):
        """ Default implementation of the coverage lookup. Simply returns all
            coverages in no specific order.
        """

        cql_text = decoder.cql
        if cql_text:
            qs = models.EOObject.objects.all()
            mapping, mapping_choices = filters.get_field_mapping_for_model(
                qs.model
            )
            ast = parse(cql_text)
            filter_expressions = to_filter(ast, mapping, mapping_choices)
            qs = qs.filter(filter_expressions)

        else:
            qs = models.EOObject.objects.filter(
                Q(
                    service_visibility__service='wcs',
                    service_visibility__visibility=True
                ) | Q(  # include mosaics with a Grid
                    mosaic__isnull=False,
                    mosaic__grid__isnull=False,
                    service_visibility__service='wcs',
                    service_visibility__visibility=True
                )
            )

        return qs.order_by(
            "identifier"
        ).select_subclasses(models.Coverage, models.Mosaic)

    def get_params(self, coverages, decoder):
        """ Default method to return a render params object from the given
            coverages/decoder.
        """

        return WCSCapabilitiesRenderParams(
            coverages,
            getattr(decoder, "version", None),
            getattr(decoder, "sections", None),
            getattr(decoder, "acceptlanguages", None),
            getattr(decoder, "acceptformats", None),
            getattr(decoder, "updatesequence", None),
        )

    def get_renderer(self, params):
        """ Default implementation for a renderer retrieval.
        """
        renderer = get_capabilities_renderer(params)
        if not renderer:
            raise OperationNotSupportedException(
                "No Capabilities renderer found for the given parameters.",
                self.request
            )
        return renderer

    def to_http_response(self, result_set):
        """ Default result to response conversion method.
        """
        return to_http_response(result_set)

    def handle(self, request):
        """ Default handler method.
        """

        # parse the parameters
        decoder = self.get_decoder(request)

        # get the coverages
        coverages = self.lookup_coverages(decoder)

        # create the render params
        params = self.get_params(coverages, decoder)
        params.http_request = request

        # get the renderer
        renderer = self.get_renderer(params)

        # dispatch the renderer and return the response
        result_set = renderer.render(params)
        return self.to_http_response(result_set)


class WCSDescribeCoverageHandlerBase(object):
    """ Base for Coverage description handlers.
    """

    service = "WCS"
    request = "DescribeCoverage"

    index = 1

    def get_decoder(self, request):
        """ Interface method to get the correct decoder for this request.
        """

    def lookup_coverages(self, decoder):
        """ Default implementation of the coverage lookup. Returns a sorted list
            of coverage models according to the decoders `coverage_ids`
            attribute. Raises a `NoSuchCoverageException` if any of the given
            IDs was not found in the database.
        """
        ids = decoder.coverage_ids

        # qs = models.Coverage.objects.filter(identifier__in=ids)
        qs = models.EOObject.objects.filter(
            identifier__in=ids,
        ).filter(
            Q(coverage__isnull=False) | Q(mosaic__isnull=False)
        ).select_subclasses()

        objects = sorted(
            qs,
            key=(lambda coverage: ids.index(coverage.identifier))
        )

        # check correct number
        if len(objects) < len(ids):
            available_ids = set([coverage.identifier for coverage in objects])
            raise NoSuchCoverageException(set(ids) - available_ids)

        return [
            Coverage.from_model(obj)
            if isinstance(obj, models.Coverage) else Mosaic.from_model(obj)
            for obj in objects
        ]

    def get_params(self, coverages, decoder):
        """ Interface method to return a render params object from the given
            coverages/decoder.
        """

    def get_renderer(self, params):
        """ Default implementation for a renderer retrieval.
        """

        renderer = get_coverage_description_renderer(params)
        if not renderer:
            raise OperationNotSupportedException(
                "No suitable coverage description renderer found.",
                self.request
            )
        return renderer

    def to_http_response(self, result_set):
        """ Default result to response conversion method.
        """
        return to_http_response(result_set)

    def handle(self, request):
        """ Default request handling method implementation.
        """
        # parse the parameters
        decoder = self.get_decoder(request)

        # lookup the coverages
        coverages = self.lookup_coverages(decoder)

        # create the render parameters
        params = self.get_params(coverages, decoder)

        # find the correct renderer
        renderer = self.get_renderer(params)

        # render and return the response
        result_set = renderer.render(params)
        return self.to_http_response(result_set)


class WCSGetCoverageHandlerBase(object):
    """ Base for get coverage handlers.
    """

    service = "WCS"
    request = "GetCoverage"

    index = 10

    def get_decoder(self, request):
        """ Interface method to get the correct decoder for this request.
        """

    def get_subsets(self, decoder):
        """ Interface method to get the subsets for this request.
        """

    def lookup_coverage(self, decoder, subsets):
        """ Default implementation of the coverage lookup. Returns the coverage
            model for the given request decoder or raises an exception if it is
            not found.
        """
        coverage_id = decoder.coverage_id

        try:
            obj = models.EOObject.objects.select_subclasses(
                models.Coverage, models.Mosaic
            ).get(
                Q(identifier=coverage_id) & (
                    Q(coverage__isnull=False) | Q(mosaic__isnull=False)
                )
            )
        except models.EOObject.DoesNotExist:
            raise NoSuchCoverageException((coverage_id,))

        if isinstance(obj, models.Coverage):
            return Coverage.from_model(obj)
        else:
            coverages = obj.coverages.all().order_by("begin_time")
            if subsets:
                subset_polygon = subsets.bounding_polygon(
                    Mosaic.from_model(obj, [])
                )
                coverages = coverages.filter(
                    footprint__intersects=subset_polygon
                )
            return Mosaic.from_model(obj, coverages)

    def get_params(self, coverages, decoder, request):
        """ Interface method to return a render params object from the given
            coverages/decoder.
        """

    def get_renderer(self, params):
        """ Default implementation for a renderer retrieval.
        """
        renderer = get_coverage_renderer(params)
        if not renderer:
            raise OperationNotSupportedException(
                "No renderer found for coverage '%s'."
                % params.coverage.identifier,
                self.request
            )
        return renderer

    def to_http_response(self, result_set):
        """ Default result to response conversion method.
        """
        return to_http_response(result_set)

    def handle(self, request):
        """ Default handling method implementation.
        """

        # parse the request
        decoder = self.get_decoder(request)

        # get the decoded subsets
        subsets = self.get_subsets(decoder)

        # get the coverage model
        coverage = self.lookup_coverage(decoder, subsets)

        # create the render params
        params = self.get_params(coverage, decoder, request)

        # get the renderer
        renderer = self.get_renderer(params)

        # render the coverage and return the response
        result_set = renderer.render(params)
        return self.to_http_response(result_set)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

try:
    from itertools import izip_longest
except ImportError:
    from itertools import zip_longest as izip_longest

from lxml.builder import ElementMaker

from eoxserver.core.util.xmltools import XMLEncoder, NameSpace, NameSpaceMap

# namespace declarations
ns_cis = NameSpace("http://www.opengis.net/gml/1.1", "cis")
ns_swe = NameSpace("http://www.opengis.net/swe/2.0", "swe")

nsmap = NameSpaceMap(ns_cis, ns_swe)

# Element factories
CIS = ElementMaker(namespace=ns_cis.uri, nsmap=nsmap)
SWE = ElementMaker(namespace=ns_swe.uri, nsmap=nsmap)


class CIS11XMLEncoder(XMLEncoder):
    def encode_axis_extent(self, axis, origin, size):
        if axis.regular:
            return CIS('axisExtent',
                axisLabel=axis.name,
                uomLabel=axis.uom,
                lowerBound=str(origin),
                upperBound=str(origin + (axis.offset * size)),
            )
        else:
            return CIS('axisExtent',
                axisLabel=axis.name,
                uomLabel=axis.uom,
                lowerBound=str(axis.positions[0]),
                upperBound=str(axis.positions[-1]),
            )

    def encode_envelope(self, grid, origins, sizes):
        return CIS('envelope', *[
                self.encode_axis_extent(axis, origin, size)
                for axis, origin, size in izip_longest(grid, origins, sizes)
            ],
            axisLabels=' '.join(axis.name for axis in grid),
            srsName=grid.coordinate_reference_system,
            srsDimension=str(len(grid))
        )

    def encode_regular_axis(self, axis, origin, size):
        """
        """
        return CIS('RegularAxis',
            axisLabel=axis.name,
            uomLabel=axis.uom,
            lowerBound=str(origin),
            upperBound=str(origin + (axis.offset * size)),
            resolution=str(axis.offset),
        )

    def encode_irregular_axis(self, axis):
        """
        """
        return CIS('IrregularAxis', *[
                CIS('C', str(position))
                for position in axis.positions
            ],
            axisLabel=axis.name,
            uomLabel=axis.uom,
        )

    def encode_grid(self, grid, origins, sizes):
        return CIS('generalGrid', *([
                self.encode_regular_axis(axis, origin, size)
                if axis.regular else
                self.encode_irregular_axis(axis)
                for axis, origin, size in izip_longest(grid, origins, sizes)
            ] + [
                CIS('gridLimits', *[
                    CIS(
                        'indexAxis',
                        axisLabel=label,
                        lowerBound="0",
                        upperBound=str(size),
                    ) for label, size in zip('ijkl', sizes)
                ])
            ]),
            srsName=grid.coordinate_reference_system,
            axisLabels=' '.join(axis.name for axis in grid)
        )

    def encode_domain_set(self, coverage):
        return CIS('domainSet',
            self.encode_grid(coverage.grid, coverage.origin, coverage.size)
        )

    def encode_nil_values(self, nil_values):
        return SWE("nilValues",
            SWE("NilValues",
                *[
                    SWE("nilValue", nil_value[0], reason=nil_value[1])
                    for nil_value in nil_values
                ]
            )
        )

    def encode_field(self, field):
        return SWE("field",
            SWE("Quantity",
                SWE("description", field.description),
                self.encode_nil_values(field.nil_values),
                SWE("uom", code=field.unit_of_measure),
                SWE("constraint",
                    SWE("AllowedValues",
                        *[
                            SWE("interval", "%g %g" % value_range)
                            for value_range in field.allowed_values
                        ] + [
                            SWE("significantFigures", str(
                                field.significant_figures
                            ))
                        ] if field.significant_figures else []
                    )
                ),
                # http://www.opengis.net/def/dataType/OGC/0/
                definition=field.definition
            ),
            name=field.identifier
        )

    def encode_range_type(self, range_type):
        return CIS('rangeType',
            SWE('DataRecord', *[
                    self.encode_field(field) for field in range_type
                ]
            )
        )

    def encode_general_grid_coverage(self, coverage):
        pass



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import os.path

from lxml.etree import parse, fromstring
from django.contrib.gis.geos import MultiPolygon, Polygon
from django.utils.six import iteritems
from eoxserver.resources.coverages import crss


try:
    import s2reader
    HAVE_S2READER = True
except ImportError:
    HAVE_S2READER = False


class S2ProductFormatReader(object):
    def test_path(self, path):
        if not HAVE_S2READER:
            return False

        try:
            with s2reader.open(path):
                pass
            return True
        except IOError:
            return False

    def read_path(self, path):
        values = {}
        with s2reader.open(path) as ds:
            metadata = ds._product_metadata
            granule = ds.granules[0]
            granule_metadata = granule._metadata

            values['identifier'] = metadata.findtext(
                './/PRODUCT_URI'
            )

            values['begin_time'] = ds.product_start_time
            values['end_time'] = ds.product_stop_time
            values['footprint'] = ds.footprint.wkt

            values['masks'] = [
                ('clouds', self._read_mask(granule, 'MSK_CLOUDS')),
                ('nodata', self._read_mask(granule, 'MSK_NODATA')),
            ]

            def tci_path(granule):
                tci_paths = [
                    path for path in granule.dataset._product_metadata.xpath(
                        ".//Granule[@granuleIdentifier='%s']/IMAGE_FILE/text()"
                        % granule.granule_identifier
                    ) if path.endswith('TCI')
                ]
                try:
                    return os.path.join(
                        ds._zip_root if ds.is_zip else ds.path,
                        tci_paths[0]
                    ) + '.jp2'
                except IndexError:
                    raise IOError(
                        "TCI path does not exist"
                    )

            values['browses'] = [
                (None, tci_path(granule))
            ]

            # TODO: extended metadata

            # values['parent_identifier']
            # values['production_status']
            # values['acquisition_type']
            values['orbit_number'] = ds.sensing_orbit_number
            values['orbit_direction'] = ds.sensing_orbit_direction
            # values['track']
            # values['frame']
            values['swath_identifier'] = metadata.find('.//Product_Info/Datatake').attrib['datatakeIdentifier']
            values['product_version'] = metadata.findtext('.//Product_Info/PROCESSING_BASELINE')
            # values['product_quality_status']
            # values['product_quality_degradation_tag']
            # values['processor_name']
            # values['processing_center']
            # values['creation_date']
            # values['modification_date']
            values['processing_date'] = ds.generation_time
            # values['sensor_mode']
            values['archiving_center'] = granule_metadata.findtext('.//ARCHIVING_CENTRE')
            # values['processing_mode']

            values['availability_time'] = ds.generation_time
            # values['acquisition_station']
            # values['acquisition_sub_type']
            # values['start_time_from_ascending_node']
            # values['completion_time_from_ascending_node']
            values['illumination_azimuth_angle'] = metadata.findtext('.//Mean_Sun_Angle/AZIMUTH_ANGLE')
            values['illumination_zenith_angle'] = metadata.findtext('.//Mean_Sun_Angle/ZENITH_ANGLE')
            # values['illumination_elevation_angle']
            # values['polarisation_mode']
            # values['polarization_channels']
            # values['antenna_look_direction']
            # values['minimum_incidence_angle']
            # values['maximum_incidence_angle']

            # values['doppler_frequency']
            # values['incidence_angle_variation']

            values['cloud_cover'] = metadata.findtext(".//Cloud_Coverage_Assessment")
            # values['snow_cover']
            # values['lowest_location']
            # values['highest_location']

        return values

    def _read_mask(self, granule, mask_type):
        for item in next(granule._metadata.iter("Pixel_Level_QI")):
            if item.attrib.get("type") == mask_type:
                gml_filename = os.path.join(
                    granule.granule_path, "QI_DATA", os.path.basename(item.text)
                )

        if granule.dataset.is_zip:
            root = fromstring(granule.dataset._zipfile.read(gml_filename))
        else:
            root = parse(gml_filename).getroot()
        return parse_mask(root)


def parse_mask(mask_elem):
    nsmap = {k: v for k, v in iteritems(mask_elem.nsmap) if k}
    # name = mask_elem.xpath('gml:name/text()', namespaces=nsmap)[0]
    try:
        crs = mask_elem.xpath(
            'gml:boundedBy/gml:Envelope/@srsName', namespaces=nsmap
        )[0]
    except IndexError:
        # just return an empty polygon when no mask available
        return MultiPolygon()

    srid = crss.parseEPSGCode(crs, [crss.fromURN])
    swap = crss.hasSwappedAxes(srid)

    mask_features = [
        parse_polygon(polygon_elem, nsmap, swap)
        for polygon_elem in mask_elem.xpath(
            'eop:maskMembers/eop:MaskFeature/eop:extentOf/gml:Polygon',
            namespaces=nsmap
        )
    ]
    return MultiPolygon(mask_features, srid=srid)


def parse_polygon(polygon_elem, nsmap, swap_axes):
    return Polygon(*[
            parse_pos_list(
                polygon_elem.xpath(
                    'gml:exterior/gml:LinearRing/gml:posList', namespaces=nsmap
                )[0], swap_axes
            )
        ] + [
            parse_pos_list(pos_list_elem, swap_axes)
            for pos_list_elem in polygon_elem.xpath(
                'gml:interior/gml:LinearRing/gml:posList', namespaces=nsmap
            )
        ]
    )


def parse_pos_list(pos_list_elem, swap_axes):
    # retrieve the number of elements per point
    dims = int(pos_list_elem.attrib.get('srsDimension', '2'))
    parts = [float(coord) for coord in pos_list_elem.text.strip().split()]

    ring = []
    i = 0
    while i < len(parts):
        ring.append(
            (parts[i + 1], parts[i]) if swap_axes else (parts[i], parts[i + 1])
        )
        i += dims

    return ring



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import os
import os.path
try:
    from io import StringIO
except ImportError:
    from cStringIO import StringIO

from uuid import uuid4

from django.http import HttpResponse
from django.utils.six import b

from eoxserver.core.util import multiparttools as mp


class ResultItem(object):
    """ Base class (or interface) for result items of a result set.

    :param content_type: the content type of the result item. in HTTP this will
                         be translated to the ``Content-Type`` header
    :param filename: the filename of the result item.
    :param identifier: the identifier of the result item. translated to
                       ``Content-Id`` HTTP header
    """

    def __init__(self, content_type=None, filename=None, identifier=None):
        self._content_type = content_type
        self.filename = filename
        self.identifier = identifier

    @property
    def data(self):
        """ Returns the "raw" data, usually as a string, buffer, memoryview,
        etc.
        """
        return ""

    @property
    def data_file(self):
        """ Returns the data as a Python file-like object.
        """
        return StringIO("")

    @property
    def content_type(self):
        """ Reterns a binary value of content-type if it is a string. 
        """
        if isinstance(self._content_type, str):
            self._content_type = b(self._content_type)
        return self._content_type

    def __len__(self):
        """ Unified access to size of data.
        """
        raise NotImplementedError

    size = property(lambda self: len(self))

    def chunked(self, chunksize):
        """ Returns a chunk of the data, which has at most ``chunksize`` bytes.
        """
        yield ""

    def delete(self):
        """ Cleanup any associated files, allocated memory, etc.
        """
        pass


class ResultFile(ResultItem):
    """ Class for results that wrap physical files on the disc.
    """

    def __init__(self, path, content_type=None, filename=None, identifier=None):
        super(ResultFile, self).__init__(content_type, filename, identifier)
        self.path = path

    @property
    def data(self):
        with open(self.path, 'rb') as f:
            return f.read()

    @property
    def data_file(self):
        return open(self.path)

    def __len__(self):
        return os.path.getsize(self.path)

    def chunked(self, chunksize):
        with open(self.path, 'rb') as f:
            while True:
                data = f.read(chunksize)
                if not data:
                    break

                yield data

    def delete(self):
        os.remove(self.path)


class ResultBuffer(ResultItem):
    """ Class for results that are actually a subset of a larger context.
        Usually a buffer.
    """

    def __init__(self, buf, content_type=None, filename=None, identifier=None):
        super(ResultBuffer, self).__init__(content_type, filename, identifier)
        if isinstance(buf, str) or isinstance(buf, bytes):
            self.buf = buf
        else: 
            self.buf = buf.tobytes()

    @property
    def data(self):
        return self.buf
    # @property
    # def data_file(self):
    #     return StringIO(self.buf)

    def __len__(self):
        return len(self.buf)

    def chunked(self, chunksize):
        if chunksize < 0:
            raise ValueError

        size = len(self.buf)
        if chunksize >= size:
            yield self.buf
            return

        i = 0
        while i < size:
            yield self.buf[i:i+chunksize]
            i += chunksize


def get_content_type(result_set):
    """ Returns the content type of a result set. If only one item is included
        its content type is used, otherwise the constant "multipart/related".
    """
    if len(result_set) == 1:
        return result_set[0].content_type
    else:
        return "multipart/related"


def get_headers(result_item):
    """ Yields content headers, if they are set in the result item.
    """
    if hasattr(result_item, 'headers'):
        for header in result_item.headers:
            yield header
    yield b"Content-Type", result_item.content_type or b"application/octet-stream"
    if result_item.identifier:
        yield b"Content-Id", result_item.identifier.encode('utf-8')
        if  isinstance(result_item.filename, str):
            result_item.filename = b(result_item.filename)
    if result_item.filename:
        yield (
            b"Content-Disposition", b'attachment; filename="%s"'
            % result_item.filename
        )
    try:
        yield b"Content-Length", b"%d" % len(result_item)
    except (AttributeError, NotImplementedError):
        pass


def get_payload_size(result_set, boundary):
    """ Calculate the size of the result set and all entailed result items plus
        headers.
    """
    boundary_str = b"%s--%s%s" % (mp.CRLF, boundary, mp.CRLF)
    boundary_str_end = b"%s--%s--" % (mp.CRLF, boundary)

    size = 0
    for item in result_set:
        size += len(boundary_str)
        size += len(
            mp.CRLF.join(b"%s: %s" % (k, v) for k, v in get_headers(item))
        )
        size += len(mp.CRLFCRLF)
        size += len(item)
    size += len(boundary_str_end)
    return size


def to_http_response(result_set, response_type=HttpResponse, boundary=None):
    """ Returns a response for a given result set. The ``response_type`` is the
        class to be used. It must be capable to work with iterators. This
        function is also responsible to delete any temporary files and buffers
        of the ``result_set``.

        :param result_set: an iterable of objects following the
                           :class:`ResultItem` interface
        :param response_type: the response type class to use; defaults to
                              :class:`HttpResponse <django.http.HttpResponse>`.
                              For streaming responses use
                              :class:`StreamingHttpResponse
                              <django.http.StreamingHttpResponse>`
        :param boundary: the multipart boundary; if omitted a UUID hex string is
                         computed and used
        :returns: a response object of the desired type
    """

    # if more than one item is contained in the result set, the content type is
    # multipart
    if len(result_set) > 1:
        boundary = boundary or (uuid4().hex).encode('utf-8')
        content_type = b"multipart/related; boundary=%s" % boundary
        headers = (('Content-Length', get_payload_size(result_set, boundary)), )

    # otherwise, the content type is the content type of the first included item
    elif len(result_set) < 1 or result_set[0].content_type is None:
        boundary = None
        content_type = b"application/octet-stream"
        headers = (('Content-Length', 0 ),)

    else:
        boundary = None
        content_type = result_set[0].content_type or b"application/octet-stream"
        headers = tuple(get_headers(result_set[0]))

    def response_iterator(items, boundary=None):
        try:
            if boundary:
                boundary_str = b"%s--%s%s" % (mp.CRLF, boundary, mp.CRLF)
                boundary_str_end = b"%s--%s--" % (mp.CRLF, boundary)

            for item in items:
                if boundary:
                    yield boundary_str
                    yield mp.CRLF.join(
                    b"%s: %s" % (key, value) 
                    for key, value in get_headers(item)
                    ) + mp.CRLFCRLF
                yield item.data
            if boundary:
                yield boundary_str_end
        finally:
            for item in items:
                try:
                    item.delete()
                except:
                    pass  # bad exception swallowing...

    # workaround for bug in django, that does not consume iterator in tests.
    if response_type == HttpResponse:
        response = response_type(
            list(response_iterator(result_set, boundary)),
            content_type
        )
    else:
        response = response_type(
            response_iterator(result_set, boundary), content_type
        )

    # set any headers
    for key, value in headers:
        response[key] = value

    return response


def parse_headers(headers):
    """ Convenience function to read the "Content-Type", "Content-Disposition"
        and "Content-Id" headers.

        :param headers: the raw header :class:`dict`
    """
    content_type = headers.get(b"Content-Type", b"application/octet-stream")
    _, params = mp.parse_parametrized_option(
        headers.get(b"Content-Disposition", b"")
    )
    filename = params.get(b"filename")
    if filename:
        if filename.startswith(b'"'):
            filename = filename[1:]
        if filename.endswith(b'"'):
            filename = filename[:-1]

    identifier = headers.get(b"Content-Id")
    return content_type, filename, identifier


def result_set_from_raw_data(data):
    """ Create a result set from raw HTTP data. This can either be a single
        or a multipart string. It returns a list containing objects of the
        :class:`ResultBuffer` type that reference substrings of the given data.

        :param data: the raw byte data
        :returns: a result set: a list containing :class:`ResultBuffer`
    """
    return [
        ResultBuffer(d, *parse_headers(headers))
        for headers, d in mp.iterate(data)
        if not headers.get(b"Content-Type").startswith(b"multipart")
    ]



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from pycql import parse, get_repr
from pycql.integrations.django import to_filter


from .filters import get_field_mapping_for_model


def apply(qs, cql, exclude=False):
    """ Applies a given CQL filter on a passed queryset. The field mapping is
        deducted from the model of the passed queryset.
        A new queryset is returned with all filters applied.
        :param qs: the base query to apply the filters on. The :attr:`model`
                   is used to determine the metadata field mappings.
        :param cql: a string containing the CQL expressions to be parsed and
                    applied
        :param exclude: whether the filters shall be applied using
                        :meth:`exclude`. Default is ``False``.
        :returns: A new queryset object representing the filtered queryset.
        :rtype: :class:`django.db.models.QuerySet`
    """
    mapping, mapping_choices = get_field_mapping_for_model(qs.model)
    ast = parse(cql)

    filters = to_filter(ast, mapping, mapping_choices)
    if exclude:
        return qs.exclude(filters)
    else:
        return qs.filter(filters)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.contrib import ogr
from eoxserver.contrib import mapserver as ms
from eoxserver.backends.access import get_vsi_path


class PolygonMaskConnector(object):
    """ Connects polygon mask files to MapServer polygon layers. For some
        purposes this can also be done via "reverse" polygons, where the actual
        polygons are subtracted from the coverages footprint.
    """

    def supports(self, coverage, data_items):
        num = len(data_items)
        return (
            len(data_items) >= 1 and
            len(filter(
                lambda d: d.semantic.startswith("polygonmask"), data_items
            )) == num
        )

    def connect(self, coverage, data_items, layer, options):
        mask_item = data_items[0]

        try:
            is_reversed = (
                layer.metadata.get("eoxs_geometry_reversed") == "true"
            )
        except ms.MapServerError:
            is_reversed = False

        # check if the geometry is "reversed"
        if is_reversed:
            # TODO: better use the coverages Extent?
            geom_types = (ogr.wkbPolygon, ogr.wkbMultiPolygon)
            output_polygon = ogr.Geometry(wkt=str(coverage.footprint.wkt))

            for mask_item in data_items:
                ds = ogr.Open(get_vsi_path(mask_item))
                for i in range(ds.GetLayerCount()):
                    ogr_layer = ds.GetLayer(i)
                    if not ogr_layer:
                        continue

                    feature = ogr_layer.GetNextFeature()
                    while feature:
                        # TODO: reproject if necessary
                        geometry = feature.GetGeometryRef()
                        if geometry.GetGeometryType() not in geom_types:
                            continue
                        if geometry:
                            output_polygon = output_polygon.Difference(geometry)
                        feature = ogr_layer.GetNextFeature()

            # since we have the geometry already in memory, add it to the layer
            # as WKT
            shape = ms.shapeObj.fromWKT(output_polygon.ExportToWkt())
            shape.initValues(1)
            shape.setValue(0, coverage.identifier)
            layer.addFeature(shape)

        else:
            layer.connectiontype = ms.MS_OGR
            layer.connection = connect(data_items[0])
            # TODO: more than one mask_item?

        layer.setProjection("EPSG:4326")
        layer.setMetaData("ows_srs", "EPSG:4326")
        layer.setMetaData("wms_srs", "EPSG:4326")

    def disconnect(self, coverage, data_items, layer, options):
        pass



#!/usr/bin/env python
#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from os.path import splitext
try:
    from itertools import chain, izip
except ImportError :
    from itertools import chain
    izip = zip
import numpy
import logging

from django.contrib.gis.geos import (
    GEOSGeometry, MultiPolygon, Polygon, LinearRing
)

from eoxserver.contrib import gdal, ogr, osr
from eoxserver.core.util.xmltools import XMLEncoder
from eoxserver.processing.preprocessing.util import (
    create_mem, create_mem_copy, copy_projection, cleanup_temp
)

from eoxserver.processing.preprocessing.optimization import (
    BandSelectionOptimization, ColorIndexOptimization, NoDataValueOptimization,
    OverviewOptimization, ReprojectionOptimization, AlphaBandOptimization
)


logger = logging.getLogger(__name__)


def pairwise(iterable):
    "s -> (s0,s1), (s2,s3), (s4, s5), ..."
    a = iter(iterable)
    return izip(a, a)


class NativeMetadataFormatEncoder(XMLEncoder):
    """
    Encodes EO Coverage metadata
    """

    def encodeMetadata(self, eoid, begin_time, end_time, polygon):
        return self._makeElement("", "Metadata", [
            ("", "EOID", eoid),
            ("", "BeginTime", begin_time),
            ("", "EndTime", end_time),
            ("", "Footprint", [
                ("", "Polygon", [
                    ("", "Exterior", self._posListToString(polygon[0]))
                ] + [
                    tuple("", "Interior", self._posListToString(interior))
                    for interior in polygon[1:]
                ])
            ])
        ])

    def _posListToString(self, ring):
        return " ".join(map(str, ring))

# enum for bandmode
RGB, RGBA, ORIG_BANDS = range(3)

#===============================================================================
# Pre-Processors
#===============================================================================


class PreProcessor(object):
    """
    """

    force = False

    def __init__(self, format_selection, overviews=True, crs=None, bands=None,
                 bandmode=RGB, footprint_alpha=False,
                 color_index=False, palette_file=None, no_data_value=None,
                 overview_resampling=None, overview_levels=None,
                 overview_minsize=None, radiometric_interval_min=None,
                 radiometric_interval_max=None, simplification_factor=None,
                 temporary_directory=None):

        self.format_selection = format_selection
        self.overviews = overviews
        self.overview_resampling = overview_resampling
        self.overview_levels = overview_levels
        self.overview_minsize = overview_minsize

        self.crs = crs

        self.bands = bands
        self.bandmode = bandmode
        self.footprint_alpha = footprint_alpha
        self.color_index = color_index
        self.palette_file = palette_file
        self.no_data_value = no_data_value
        self.radiometric_interval_min = radiometric_interval_min
        self.radiometric_interval_max = radiometric_interval_max

        if simplification_factor is not None:
            self.simplification_factor = simplification_factor
        else:
            # default 2 * resolution == 2 pixels
            self.simplification_factor = 2

        self.temporary_directory = temporary_directory

    def process(self, input_filename, output_filename,
                geo_reference=None, generate_metadata=True):

        # open the dataset and create an In-Memory Dataset as copy
        # to perform optimizations
        ds = create_mem_copy(gdal.Open(input_filename))

        gt = ds.GetGeoTransform()
        footprint_wkt = None

        if not geo_reference:
            if gt == (0.0, 1.0, 0.0, 0.0, 0.0, 1.0):
                # TODO: maybe use a better check
                raise ValueError("No geospatial reference for unreferenced "
                                 "dataset given.")
        else:
            logger.debug("Applying geo reference '%s'."
                         % type(geo_reference).__name__)
            ds, footprint_wkt = geo_reference.apply(ds)

        # apply optimizations
        for optimization in self.get_optimizations(ds):
            logger.debug("Applying optimization '%s'."
                         % type(optimization).__name__)

            try:
                new_ds = optimization(ds)

                if new_ds is not ds:
                    # cleanup afterwards
                    cleanup_temp(ds)
                    ds = new_ds
            except:
                cleanup_temp(ds)
                raise

        # generate the footprint from the dataset
        if not footprint_wkt:
            logger.debug("Generating footprint.")
            footprint_wkt = self._generate_footprint_wkt(ds)
        # check that footprint is inside of extent of generated image
        # regenerate otherwise
        else:
            tmp_extent = getExtentFromRectifiedDS(ds)
            tmp_bbox = Polygon.from_bbox((tmp_extent[0], tmp_extent[1],
                                          tmp_extent[2], tmp_extent[3]))
            tmp_footprint = GEOSGeometry(footprint_wkt)
            if not tmp_bbox.contains(tmp_footprint):
                footprint_wkt = tmp_footprint.intersection(tmp_bbox).wkt

        if self.footprint_alpha:
            logger.debug("Applying optimization 'AlphaBandOptimization'.")
            opt = AlphaBandOptimization()
            opt(ds, footprint_wkt)

        output_filename = self.generate_filename(output_filename)

        logger.debug("Writing file to disc using options: %s."
                     % ", ".join(self.format_selection.creation_options))

        logger.debug("Metadata tags to be written: %s"
                     % ", ".join(ds.GetMetadata_List("") or []))

        # save the file to the disc
        driver = gdal.GetDriverByName(self.format_selection.driver_name)
        ds = driver.CreateCopy(output_filename, ds,
                               options=self.format_selection.creation_options)

        for optimization in self.get_post_optimizations(ds):
            logger.debug("Applying post-optimization '%s'."
                         % type(optimization).__name__)
            optimization(ds)

        # generate metadata if requested
        footprint = None
        if generate_metadata:
            normalized_space = Polygon.from_bbox((-180, -90, 180, 90))
            non_normalized_space = Polygon.from_bbox((180, -90, 360, 90))

            footprint = GEOSGeometry(footprint_wkt)
            #.intersection(normalized_space)
            outer = non_normalized_space.intersection(footprint)

            if len(outer):
                footprint = MultiPolygon(
                    *map(lambda p:
                        Polygon(*map(lambda ls:
                            LinearRing(*map(lambda point:
                                (point[0] - 360, point[1]), ls.coords
                            )), tuple(p)
                        )), (outer,)
                    )
                ).union(normalized_space.intersection(footprint))
            else:
                if isinstance(footprint, Polygon):
                    footprint = MultiPolygon(footprint)

            logger.info("Calculated Footprint: '%s'" % footprint.wkt)

            # use the provided footprint
            #geom = OGRGeometry(footprint_wkt)
            #exterior = []
            #for x, y in geom.exterior_ring.tuple:
            #    exterior.append(y); exterior.append(x)

            #polygon = [exterior]
        num_bands = ds.RasterCount

        # finally close the dataset and write it to the disc
        ds = None

        return PreProcessResult(output_filename, footprint, num_bands)

    def generate_filename(self, filename):
        """ Adjust the filename with the correct extension. """
        base_filename, _ = splitext(filename)
        return base_filename + self.format_selection.extension

    def _generate_footprint_wkt(self, ds):
        """ Generate a footprint from a raster, using black/no-data as exclusion
        """

        # create an empty boolean array initialized as 'False' to store where
        # values exist as a mask array.
        nodata_map = numpy.zeros((ds.RasterYSize, ds.RasterXSize),
                                 dtype=numpy.bool)

        for idx in range(1, ds.RasterCount + 1):
            band = ds.GetRasterBand(idx)
            raster_data = band.ReadAsArray()
            nodata = band.GetNoDataValue()

            if nodata is None:
                nodata = 0

            # apply the output to the map
            nodata_map |= (raster_data != nodata)

        # create a temporary in-memory dataset and write the nodata mask
        # into its single band
        tmp_ds = create_mem(ds.RasterXSize + 2, ds.RasterYSize + 2, 1,
                            gdal.GDT_Byte)
        copy_projection(ds, tmp_ds)
        tmp_band = tmp_ds.GetRasterBand(1)
        tmp_band.WriteArray(nodata_map.astype(numpy.uint8))

        # create an OGR in memory layer to hold the created polygon
        sr = osr.SpatialReference()
        sr.ImportFromWkt(ds.GetProjectionRef())
        ogr_ds = ogr.GetDriverByName('Memory').CreateDataSource('out')
        layer = ogr_ds.CreateLayer('poly', sr.sr, ogr.wkbPolygon)
        fd = ogr.FieldDefn('DN', ogr.OFTInteger)
        layer.CreateField(fd)

        # polygonize the mask band and store the result in the OGR layer
        gdal.Polygonize(tmp_band, tmp_band, layer, 0)

        if layer.GetFeatureCount() > 1:
            # if there is more than one polygon, compute the minimum
            # bounding polygon
            geometry = ogr.Geometry(ogr.wkbPolygon)
            while True:
                feature = layer.GetNextFeature()
                if not feature:
                    break
                geometry = geometry.Union(feature.GetGeometryRef())

            # TODO: improve this for a better minimum bounding polygon
            geometry = geometry.ConvexHull()

        elif layer.GetFeatureCount() < 1:
            # there was an error during polygonization
            raise RuntimeError("Error during polygonization. No feature "
                               "obtained.")
        else:
            # obtain geometry from the first (and only) layer
            feature = layer.GetNextFeature()
            geometry = feature.GetGeometryRef()

        if geometry.GetGeometryType() != ogr.wkbPolygon:
            raise RuntimeError("Error during polygonization. Wrong geometry "
                               "type: %s" % ogr.GeometryTypeToName(
                                    geometry.GetGeometryType()))

        # check if reprojection to latlon is necessary
        if not sr.IsGeographic():
            dst_sr = osr.SpatialReference()
            dst_sr.ImportFromEPSG(4326)
            try:
                geometry.TransformTo(dst_sr.sr)
            except RuntimeError:
                geometry.Transform(osr.CoordinateTransformation(sr.sr, dst_sr.sr))
        
        gt = ds.GetGeoTransform()
        resolution = min(abs(gt[1]), abs(gt[5]))

        simplification_value = self.simplification_factor * resolution

        # simplify the polygon. the tolerance value is *really* vague
        try:
            # SimplifyPreserveTopology() available since OGR 1.9.0
            geometry = geometry.SimplifyPreserveTopology(simplification_value)
        except AttributeError:
            # use GeoDjango bindings if OGR is too old
            geometry = ogr.CreateGeometryFromWkt(
                GEOSGeometry(geometry.ExportToWkt()).simplify(
                    simplification_value, True
                ).wkt
            )

        return geometry.ExportToWkt()


class WMSPreProcessor(PreProcessor):
    """

        >>> prep = WMSPreProcessor(...)
        >>> prep.process(input_filename, output_filename, generate_metadata)
    """

    def get_optimizations(self, ds):
        if self.no_data_value is not None:
            yield NoDataValueOptimization(self.no_data_value)

        if self.crs:
            yield ReprojectionOptimization(self.crs, self.temporary_directory)

        if self.bandmode not in (RGB, RGBA, ORIG_BANDS):
            raise ValueError

        if self.radiometric_interval_min is not None:
            rad_min = self.radiometric_interval_min
        else:
            rad_min = "min"
        if self.radiometric_interval_max is not None:
            rad_max = self.radiometric_interval_max
        else:
            rad_max = "max"

        # if RGB is requested, use the given bands or the first 3 bands as RGB
        if self.bandmode == RGB:
            if self.bands and len(self.bands) != 3:
                raise ValueError("Wrong number of bands given. Expected 3, got "
                                 "%d." % len(self.bands))

            if ds.RasterCount == 1:
                yield BandSelectionOptimization(
                    self.bands or [(1, rad_min, rad_max),
                                   (1, rad_min, rad_max),
                                   (1, rad_min, rad_max)],
                    temporary_directory=self.temporary_directory)
            else:
                yield BandSelectionOptimization(
                    self.bands or [(1, rad_min, rad_max),
                                   (2, rad_min, rad_max),
                                   (3, rad_min, rad_max)],
                    temporary_directory=self.temporary_directory)

        # if RGBA is requested, use the given bands or the first 4 bands as RGBA
        elif self.bandmode == RGBA:
            if self.bands and len(self.bands) != 4:
                raise ValueError("Wrong number of bands given. Expected 4, got "
                                 "%d." % len(self.bands))
            if ds.RasterCount == 1:
                yield BandSelectionOptimization(
                    self.bands or [(1, rad_min, rad_max),
                                   (1, rad_min, rad_max),
                                   (1, rad_min, rad_max),
                                   (0, 0, 0)],
                    temporary_directory=self.temporary_directory)
            else:
                yield BandSelectionOptimization(
                    self.bands or [(1, rad_min, rad_max),
                                   (2, rad_min, rad_max),
                                   (3, rad_min, rad_max),
                                   (4, rad_min, rad_max)],
                    temporary_directory=self.temporary_directory)

        # when band mode is set to original bands, don't use this optimization
        elif self.bandmode == ORIG_BANDS:
            if self.bands:
                raise ValueError("Bandmode is set to 'original', but bands are "
                                 "given.")

        else:
            raise ValueError("Illegal bandmode given.")

        if self.color_index:
            yield ColorIndexOptimization(
                self.palette_file, self.temporary_directory
            )

    def get_post_optimizations(self, ds):
        if self.overviews:
            yield OverviewOptimization(self.overview_resampling,
                                       self.overview_levels,
                                       self.overview_minsize)


#===============================================================================
# PreProcess result
#===============================================================================

class PreProcessResult(object):
    """ Result storage for preprocessed datasets. """
    def __init__(self, output_filename, footprint, num_bands):
        self.output_filename = output_filename
        self._footprint = footprint
        self.num_bands = num_bands

    @property
    def footprint_raw(self):
        pass  # TODO: return as a list of tuples

    @property
    def footprint_wkt(self):
        """ Returns the stored footprint as WKT."""
        return self.footprint_geom.wkt

    @property
    def footprint_geom(self):
        """ Returns the polygon as a GEOSGeometry. """
        #return Polygon([LinearRing([
        #    Point(x, y) for y, x in pairwise(ring)
        #]) for ring in self._footprint][0])
        return self._footprint



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2019-11-13 14:23
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('coverages', '0005_eo_object_identifier_validator'),
    ]

    operations = [
        migrations.AddField(
            model_name='masktype',
            name='validity',
            field=models.BooleanField(default=False),
        ),
    ]



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage collections. This command uses sub-commands for the
        specific tasks: create, delete, insert, exclude, purge.
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create')
        delete_parser = self.add_subparser(parser, 'delete')
        insert_parser = self.add_subparser(parser, 'insert')
        exclude_parser = self.add_subparser(parser, 'exclude')
        purge_parser = self.add_subparser(parser, 'purge')
        summary_parser = self.add_subparser(parser, 'summary')
        parsers = [
            create_parser, insert_parser, exclude_parser,
            purge_parser, summary_parser
        ]

        # identifier is a common argument (except for delete it is optional,
        # if --all is tagged)
        for parser in parsers:
            parser.add_argument(
                'identifier', nargs=1, help='The collection identifier'
            )

        create_parser.add_argument(
            '--type', '-t', dest='type_name',
            help='The collection type name. Optional.'
        )
        create_parser.add_argument(
            '--grid', '-g', dest='grid_name', default=None,
            help='The optional grid name.'
        )
        create_parser.add_argument(
            '--set', '-s', dest='set_overrides',
            nargs=2, default=[], action='append',
            help=(
                'Set (or override) additional metadata tags like '
                '"platform".'
            )
        )
        delete_parser.add_argument(
            '--all', '-a', action="store_true",
            default=False, dest='all_collections',
            help=(
                'When this flag is set, all the collections are '
                'selected to be derigesterd'
            )
        )

        delete_parser.add_argument(
                'identifier', default=None, nargs='?',
                help='The identifier of the collection to delete.'
        )
        # common arguments for insertion/exclusion
        insert_parser.add_argument(
            'object_identifiers', nargs='+',
            help=(
                'The identifiers of the objects (Product or Coverage) '
                'to insert'
            )
        )
        insert_parser.add_argument(
            '--use-extent', action='store_true', default=False,
            help=(
                'Whether to simply collection the bounding box of the '
                'footprint as the collections footprint'
            )
        )
        exclude_parser.add_argument(
            'object_identifiers', nargs='+',
            help=(
                'The identifiers of the objects (Product or Coverage) '
                'to exclude'
            )
        )
        exclude_parser.add_argument(
            '--use-extent', action='store_true', default=False,
            help=(
                'Whether to simply collection the bounding box of the '
                'footprint as the collections footprint'
            )
        )

        summary_parser.add_argument(
            '--products', action='store_true', default=True,
            dest='product_summary',
            help=('Collect summary product metadata. Default.')
        )
        summary_parser.add_argument(
            '--no-products', action='store_false', default=True,
            dest='coverage_summary',
            help=("Don't collect summary product metadata.")
        )

        summary_parser.add_argument(
            '--coverages', action='store_true', default=True,
            dest='product_summary',
            help=('Collect summary coverage metadata. Default.')
        )
        summary_parser.add_argument(
            '--no-coverages', action='store_false', default=True,
            dest='coverage_summary',
            help=("Don't collect summary coverage metadata.")
        )

    @transaction.atomic
    def handle(self, subcommand, identifier, *args, **kwargs):
        """ Dispatch sub-commands: create, delete, insert, exclude, purge.
        """
        if subcommand == "create":
            self.handle_create(identifier[0], *args, **kwargs)
        elif subcommand == "delete":
            self.handle_delete(identifier, *args, **kwargs)
        elif subcommand == "insert":
            self.handle_insert(identifier[0], *args, **kwargs)
        elif subcommand == "exclude":
            self.handle_exclude(identifier[0], *args, **kwargs)
        elif subcommand == "purge":
            self.handle_purge(identifier[0], *args, **kwargs)
        elif subcommand == "summary":
            self.handle_summary(identifier[0], *args, **kwargs)

    def handle_create(self, identifier, type_name, grid_name, **kwargs):
        """ Handle the creation of a new collection.
        """
        if grid_name:
            try:
                grid = models.Grid.objects.get(name=grid_name)
            except models.Grid.DoesNotExist:
                raise CommandError("Grid %r does not exist." % grid_name)
        else:
            grid = None

        collection_type = None
        if type_name:
            try:
                collection_type = models.CollectionType.objects.get(
                    name=type_name
                )
            except models.CollectionType.DoesNotExist:
                raise CommandError(
                    "Collection type %r does not exist." % type_name
                )

        models.Collection.objects.create(
            identifier=identifier,
            collection_type=collection_type, grid=grid
        )

        print('Successfully created collection %r' % identifier)

    def handle_delete(self, identifier, all_collections, *args, **kwargs):
        """ Handle the deletion of a collection
        """
        if not all_collections and not identifier:
            raise CommandError('please specify a collection/s to remove')
        else:
            if all_collections:
                collections = models.Collection.objects.all()
            elif identifier:
                collections = [self.get_collection(identifier)]
            for collection in collections:
                try:
                    collection_id = collection.identifier
                    collection.delete()
                    self.print_msg(
                        'Successfully deregistered collection %r'
                        % collection_id
                    )
                except models.Collection.DoesNotExist:
                    raise CommandError('No such Collection %r' % identifier)

    def handle_insert(self, identifier, object_identifiers, **kwargs):
        """ Handle the insertion of arbitrary objects into a collection
        """
        collection = self.get_collection(identifier)

        objects = list(
            models.EOObject.objects.filter(
                identifier__in=object_identifiers
            ).select_subclasses()
        )

        if len(objects) != len(set(object_identifiers)):
            actual = set(obj.identifier for obj in objects)
            missing = set(object_identifiers) - actual
            raise CommandError(
                "No such object with ID%s: %s"
                % (len(missing) > 1, ", ".join(missing))
            )

        for eo_object in objects:
            try:
                models.collection_insert_eo_object(
                    collection, eo_object, kwargs.get('use_extent', False)
                )
            except Exception as e:
                raise CommandError(
                    "Could not insert object %r into collection %r. "
                    "Error was: %s"
                    % (eo_object.identifier, collection.identifier, e)
                )

            print(
                'Successfully inserted object %r into collection %r'
                % (eo_object.identifier, collection.identifier)
            )

    def handle_exclude(self, identifier, object_identifiers, **kwargs):
        """ Handle the exclusion of arbitrary objects from a collection
        """
        collection = self.get_collection(identifier)

        objects = list(
            models.EOObject.objects.filter(
                identifier__in=object_identifiers
            ).select_subclasses()
        )

        if len(objects) != len(set(object_identifiers)):
            actual = set(obj.identifier for obj in objects)
            missing = set(object_identifiers) - actual
            raise CommandError(
                "No such object with ID%s: %s"
                % (len(missing) > 1, ", ".join(missing))
            )

        for eo_object in objects:
            try:
                models.collection_exclude_eo_object(
                    collection, eo_object, kwargs.get('use_extent', False)
                )
            except Exception as e:
                raise CommandError(
                    "Could not exclude object %r from collection %r. "
                    "Error was: %s"
                    % (eo_object.identifier, collection.identifier, e)
                )

            print(
                'Successfully excluded object %r from collection %r'
                % (eo_object.identifier, collection.identifier)
            )

    def handle_purge(self, identifier, **kwargs):
        # TODO: implement
        raise CommandError(
            "Could not exclude purge collection %r: not implemented"
            % identifier
        )
        print('Successfully purged collection %r' % identifier)

    def handle_summary(self, identifier, product_summary, coverage_summary,
                       **kwargs):
        models.collection_collect_metadata(
            self.get_collection(identifier),
            False, False, False, product_summary, coverage_summary
        )
        print('Successfully collected metadata for collection %r' % identifier)

    def get_collection(self, identifier):
        """ Helper method to get a collection by identifier or raise a
            CommandError.
        """
        try:
            return models.Collection.objects.get(identifier=identifier)
        except models.Collection.DoesNotExist:
            raise CommandError("Collection %r does not exist." % identifier)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import logging

try:
    from eoxserver.contrib.mapserver import MS_VERSION_NUM
except ImportError:
    MS_VERSION_NUM = 0

from eoxserver.core import Component, implements
from eoxserver.backends.access import connect
from eoxserver.services.mapserver.interfaces import StyleApplicatorInterface


logger = logging.getLogger(__name__)


# Versions smaller than 6.2 result in a segfault in this operation.
if MS_VERSION_NUM >= 60200:
    class SLDStyleApplicator(Component):
        implements(StyleApplicatorInterface)

        def apply(self, coverage, data_items, layer):
            sld_items = filter(lambda d: (
                d.semantic.startswith("style") and d.format.upper() == "SLD"
            ), data_items)

            for sld_item in sld_items:
                sld_filename = connect(sld_item)
                with open(sld_filename) as f:
                    layer.map.applySLD(f.read())



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""
This module provides functionality to synchronize the database models in one
collection with the contents on a storage.
"""

import logging
import re
import os.path

from eoxserver.core import env
from eoxserver.backends.access import retrieve
from eoxserver.backends import models as backends
from eoxserver.backends.component import BackendComponent
from eoxserver.backends.cache import CacheContext
from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.registration.component import (
    RegistratorComponent
)
from django.utils.six import string_types

logger = logging.getLogger(__name__)

TEMPLATE_RE = re.compile(
    r"template\[(?P<value>[a-zA-Z0-9_\[\]]+)\]", re.IGNORECASE
)
SOURCE_RE = re.compile(
    r"source\[(?P<value>[a-zA-Z0-9_\[\]]+)\]", re.IGNORECASE
)


class SynchronizationError(Exception):
    pass


def synchronize(collection, recursive=False, force=False):
    """ Synchronizes a :class:`eoxserver.resources.coverages.models.Collection`
        and all its data sources with their respective storages. Synchronization
        means to compare the contents of a collection with the contents of the
        file system referenced by each of the datasources.

        :param collection: either the ID of a collection or the collection model
                           itself. Internally the collection will always be
                           cast to its actual type.
        :param recursive: perform a recursive synchronization for all
                          sub-collections of the specified one. By default this
                          is not performed.
        :param force: force the re-registration of already inserted datasets.
                      By default, existing datasets will be preserved and not
                      refreshed.
        :returns: a two-tuple: the number of newly registered and deleted stale
                  datasets.
    """

    # allow both model and identifier
    if isinstance(collection, string_types):
        collection = models.Collection.objects.get(identifier=collection)

    collection = collection.cast()
    if models.iscoverage(collection):
        overrides = {"range_type_name": collection.range_type.name}
    else:
        overrides = {}

    if recursive:
        synchronize(
            models.Collection.objects.filter(collections__in=[collection.pk]),
            recursive
        )

    logger.info("Synchronizing collection %s" % collection)

    # expand all filesystem globs to actually existing paths
    all_paths = []
    for data_source in collection.data_sources.all():
        all_paths.extend(
            _expand_data_source(data_source)
        )

    deleted_count = 0

    with CacheContext() as cache:
        # loop over all paths and check if there is already a dataset registered
        # for it.
        registered_pks = []
        for paths in all_paths:
            for filename, data_item, semantic in paths:
                exists = backends.DataItem.objects.filter(
                    package=data_item.package, storage=data_item.storage,
                    location=filename
                ).exists()
                if not exists:
                    break

            if not exists:
                logger.info("Creating new dataset.")
                for registrator in RegistratorComponent(env).registrators:
                    # TODO: select registrator
                    pass

                # translate for registrator
                items = [
                    (d.storage or d.package, location, semantic, d.format)
                    for location, d, semantic in paths
                ]
                dataset = registrator.register(
                    items, overrides, cache
                )
                collection.insert(dataset)
                registered_pks.append(dataset.pk)

        # loop over all coverages in this collection. if any is not represented
        # by its referenced file, delete it. Skip the just added datasets for
        # convenience (as we know that they must have referenced files)

        # TODO: large exclusions like this run into problems with SQLite
        # contained = models.Coverage.objects.filter(
        #     collections__in=[collection.pk]
        # ).exclude(pk__in=registered_pks)

        # TODO: temporary (but slow) workaround
        registered_pks = set(registered_pks)
        contained = list(
            c for c in models.Coverage.objects.filter(
                collections__in=[collection.pk]
            )
            if c.pk not in registered_pks
        )

        for coverage in contained:
            data_items = tuple(coverage.data_items.all())

            # loop over all data items of the coverage and check if it was found
            # in the filesystem lookup
            for data_item in data_items:

                found = False
                # loop over all paths and check if the data item was still found
                # in the filesystem
                for paths in all_paths:
                    if found:
                        break

                    for filename, _, semantic in paths:
                        if data_item.location == filename and \
                                data_item.semantic == semantic:
                            found = True
                            break

                # if the data item as not found in the paths on the filesystem,
                # delete the model
                if not found:
                    logger.info("Deleting dataset '%s'." % coverage.identifier)
                    coverage.delete()
                    deleted_count += 1
                    break

        return len(registered_pks), deleted_count


def _expand_data_source(data_source):
    """ Helper function to loop over all files referenced by a data source.
    """

    data_items = tuple(data_source.data_items.all())

    # detect the template data items and the primary data item of a data source
    template_data_items = [
        data_item for data_item in data_items
        if TEMPLATE_RE.match(data_item.semantic)
    ]
    primary_data_items = [
        data_item for data_item in data_items
        if SOURCE_RE.match(data_item.semantic)
    ]

    if len(primary_data_items) == 0:
        raise SynchronizationError(
            "No primary data item for data source specified."
        )
    elif len(primary_data_items) > 1:
        raise SynchronizationError(
            "Too many primary data items for data source specified."
        )
    else:
        primary_data_item = primary_data_items[0]

    primary_sem = SOURCE_RE.match(primary_data_item.semantic).group("value")

    # TODO: expand data item to filenames

    files = _expand_data_item(primary_data_item)
    if not files:
        logger.info(
            "Data source %s did not match any files." % primary_data_item
        )
    else:
        logger.info(
            "Data source %s matched %d files." % (primary_data_item, len(files))
        )

    # TODO: probably replace this with a generator based solution

    all_paths = []
    for filename in files:
        # each template must match exactly one file
        dirname, basename = os.path.split(filename)
        root, ext = os.path.splitext(basename)
        template_values = {
            "basename": basename,
            "root": root,
            "dirname": dirname,
            "extension": ext,
            "source": filename
        }

        paths = [(filename, primary_data_item, primary_sem)]
        for template_data_item in template_data_items:
            template_sem = TEMPLATE_RE.match(
                template_data_item.semantic
            ).group("value")

            template_file = _expand_template_location(
                template_data_item, template_values
            )
            paths.append((template_file, template_data_item, template_sem))

        all_paths.append(paths)

    return all_paths


def _expand_data_item(data_item, cache=None):
    """ Helper function to expand a source data item to a list of file
        identifiers.
    """

    backends = BackendComponent(env)

    storage = data_item.storage
    package = data_item.package

    if storage:
        # get list of files of that storage
        component = backends.get_storage_component(storage.storage_type)
        if not component:
            raise ValueError(
                "No storage component for type '%s' found."
                % storage.storage_type
            )

        return component.list_files(storage.url, data_item.location)

    elif package:
        # get list of files of that package
        local_filename = retrieve(package, cache)
        component = backends.get_package_component(package.format)
        if not component:
            raise ValueError(
                "No package component for type '%s' found."
                % package.format
            )

        return component.list_files(local_filename, data_item.location)

    else:
        # This is a local filename, expand it directly
        component = backends.get_storage_component("local")
        if not component:
            raise ValueError("No active local storage component found.")

        return component.list_files("", data_item.location)


def _expand_template_location(data_item, template_values):
    """ Helper function to expand a source data item to a list of file
        identifiers, but first expanding the template location with the string
        format syntax.
    """

    location = data_item.location

    try:
        location = location.format(**template_values)
        # allow both formatting mechanisms
        if location == data_item.location:
            location = location % template_values
    except:
        raise ValueError(
            "Invalid template '%s' in template data item %s."
            % (location, data_item)
        )

    return location



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from functools import wraps
try:
    from functools import total_ordering
except ImportError:
    from eoxserver.core.util.functools import total_ordering

from django.utils.six import string_types


__all__ = ["parse_version_string", "Version"]


def parse_version_string(version_string):
    """ Convenience function to parse a version from a string.
    """
    return Version(*map(int, version_string.split(".")))


def convert_to_version(f):
    """ Decorator to automatically parse a ``Version`` from a string.
    """

    @wraps(f)
    def wrapper(self, other):
        if isinstance(other, Version):
            return f(self, other)
        elif isinstance(other, string_types):
            return f(self, parse_version_string(other))
        try:
            return f(self, Version(*other))
        except TypeError:
            pass
        raise TypeError("Cannot convert '%s' to version" % type(other).__name__)
    return wrapper


@total_ordering
class Version(object):
    """ Abstraction for OWS versions. Must be in the form 'x.y(.z)', where all
        components must be positive integers or zero. The last component may be
        unspecified (None).

        Versions can be compared with other versions. Strings and tuples of the
        correct layout are also compareable.

        Versions are compared by the "major" and the "minor" number. Only if
        both versions provide a "revision" it is taken into account. So Versions
        "1.0" and "1.0.1" are considered equal!
    """

    def __init__(self, major, minor, revision=None):
        try:
            assert(isinstance(major, int) and major >= 0)
            assert(isinstance(minor, int) and minor >= 0)
            assert(revision is None
                   or (isinstance(revision, int) and revision >= 0))
        except AssertionError:
            raise ValueError("Invalid version components supplied.")

        if revision is None:
            self._values = (major, minor)
        else:
            self._values = (major, minor, revision)

    @property
    def major(self):
        return self._values[0]

    @property
    def minor(self):
        return self._values[1]

    @property
    def revision(self):
        try:
            return self._values[2]
        except IndexError:
            return None

    @convert_to_version
    def __eq__(self, other):
        for self_v, other_v in zip(self._values, other._values):
            if self_v != other_v:
                return False
        return True

    @convert_to_version
    def __lt__(self, other):
        for self_v, other_v in zip(self._values, other._values):
            if self_v < other_v:
                return True
            elif self_v > other_v:
                return False
        return False

    def __str__(self):
        return ".".join(map(str, self._values))

    def __repr__(self):
        return '<%s.%s ("%s") instance at 0x%x>' % (
            __name__, type(self).__name__, str(self), id(self)
        )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

from eoxserver.contrib.vsi import open as vsi_open


def is_landsat8_l1_metadata_file(path):
    """ Checks whether the referenced file is a Landsat 8 metadata file """
    try:
        with vsi_open(path) as f:
            lines = _read_lines(f)

        return next(iter(lines)).strip() == "GROUP = L1_METADATA_FILE"
    except (ValueError, StopIteration):
        return False


def is_landsat8_l1_metadata_content(content):
    """ Checks whether the referenced file is a Landsat 8 metadata file """
    try:
        f = StringIO(content.decode())
        f.seek(0)
        return next(f).strip() == "GROUP = L1_METADATA_FILE"
    except (ValueError, StopIteration):
        return False


def parse_landsat8_l1_metadata_file(path):
    """ Parses a Landsat 8 metadata file to a nested dict representation"""
    with vsi_open(path) as f:
        lines = _read_lines(f)

    iterator = iter(lines)
    _, _ = _parse_line(next(iterator))
    return _parse_group(iterator)


def parse_landsat8_l1_metadata_content(content):
    """ Parses a Landsat 8 metadata file to a nested dict representation"""
    f = StringIO(content.decode())
    f.seek(0)
    _, _ = _parse_line(next(f))
    return _parse_group(f)


def _read_lines(f):
    return f.read().split(b'\n')


def _parse_group(iterator):
    group = {}
    for line in iterator:
        key, value = _parse_line(line)
        if not key or key == "END_GROUP":
            break
        elif key == "GROUP":
            key = value
            value = _parse_group(iterator)
        group[key] = value
    return group


def _parse_line(line):
    line = line.strip()
    if not line or line == "END":
        return (None, None)

    key, _, value = line.partition(" = ")
    if value.startswith('"') and value.endswith('"'):
        value = value[1:-1]

    return key, value



#!/usr/bin/python
#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2016 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


"""
Create a new EOxServer instance. This instance will create a root directory
with the instance name in the given (optional) directory.
"""

import sys
from optparse import OptionParser
import textwrap

from django.core.management.base import OutputWrapper
from django.core.management.color import color_style

import eoxserver
from eoxserver.core.instance import create_instance


def main():
    parser = OptionParser(
        usage=textwrap.dedent("""\
            %prog [options] <instance-name> [<target-directory>]

            Creates a new EOxServer instance with all necessary files and
            folder structure. Optionally, a SQLite database is initiated.
        """),
        version=eoxserver.get_version()
    )
    parser.add_option('-i', '--init-spatialite', '--init_spatialite',
        dest='init_spatialite', action='store_true', default=False,
        help='Flag to initialize the sqlite database.'
    )
    parser.add_option('-v', '--verbosity',
        action='store', dest='verbosity', default='1',
        type='choice', choices=['0', '1', '2', '3']
    )
    parser.add_option('--traceback',
        action='store_true', help='Raise on exception'
    )

    options, args = parser.parse_args()

    error_stream = OutputWrapper(sys.stderr, color_style().ERROR)

    if not args:
        error_stream.write("Mandatory argument 'instance-name' not given.\n")
        sys.exit(1)

    name = args[0]
    try:
        target = args[1]
    except IndexError:
        target = None

    try:
        options = options.__dict__
        options['verbosity'] = int(options['verbosity'])
        create_instance(name, target, **options)
    except Exception as e:
        if options.traceback:
            raise
        error_stream.write("%s: %s\n" % (e.__class__.__name__, e))

if __name__ == "__main__":
    main()



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from django.contrib.gis.geos import GEOSGeometry, Polygon, MultiPolygon

from eoxserver.contrib import gdal
from eoxserver.processing.gdal import reftools as rt
from eoxserver.contrib import osr
from eoxserver.resources.coverages.metadata.coverage_formats import (
    get_gdal_dataset_format_readers
)


def open_gdal(obj):
    if isinstance(obj, gdal.Dataset):
        return obj
    try:
        return gdal.Open(obj)
    except RuntimeError:
        return None


class GDALDatasetMetadataReader(object):
    def test(self, obj):
        return open_gdal(obj) is not None

    def get_format_name(self, obj):
        ds = open_gdal(obj)
        if not ds:
            return None

        driver = ds.GetDriver()
        return "GDAL/" + driver.ShortName

    def read(self, obj):
        ds = open_gdal(obj)
        if ds is None:
            raise Exception("Could not parse from obj '%s'." % repr(obj))

        driver = ds.GetDriver()
        size = (ds.RasterXSize, ds.RasterYSize)
        values = {"size": size}

        projection = ds.GetProjection()

        # --= rectified datasets =--
        # NOTE: If the projection is a non-zero string then
        #       the geocoding is given by the Geo-Trasnformation
        #       matrix - not matter what are the values.
        if projection and not (ds.GetGCPProjection() and ds.GetGCPCount() > 0):
            sr = osr.SpatialReference(projection)
            if sr.srid is not None:
                projection = 'EPSG:%d' % sr.srid

            gt = ds.GetGeoTransform()

            values['origin'] = [gt[0], gt[3]]

            values['grid'] = {
                'coordinate_reference_system': projection,
                'axis_offsets': [gt[1], gt[5]],
                'axis_types': ['spatial', 'spatial'],
                'axis_names': ['x', 'y'] if sr.IsProjected() else ['long', 'lat'],
            }

            if sr.GetLinearUnitsName() in ('metre', 'meter', 'm') \
                    and abs(gt[1]) == abs(gt[5]):
                values['grid']['resolution'] = abs(gt[1])

        # --= tie-point encoded referenceable datasets =--
        # NOTE: If the GCP projection is a non-zero string and
        #       there are GCPs we are dealing with a tie-point geocoded
        #       referenceable dataset. The extent is given by the image
        #       footprint. The fooprint must not be wrapped arround
        #       the date-line!
        elif ds.GetGCPProjection() and ds.GetGCPCount() > 0:
            projection = ds.GetGCPProjection()
            sr = osr.SpatialReference(projection)
            if sr.srid is not None:
                projection = 'EPSG:%d' % sr.srid

            values['grid'] = {
                'coordinate_reference_system': projection,
                'axis_offsets': [None, None],
                'axis_types': ['spatial', 'spatial'],
                'axis_names': ['x', 'y'] if sr.IsProjected() else ['long', 'lat']
            }
            values['origin'] = [None, None]

            # # parse the spatial reference to get the EPSG code
            sr = osr.SpatialReference(ds.GetGCPProjection(), "WKT")

            # NOTE: GeosGeometry can't handle non-EPSG geometry projections.
            if sr.GetAuthorityName(None) == "EPSG":
                srid = int(sr.GetAuthorityCode(None))

                # get the footprint
                rt_prm = rt.suggest_transformer(ds)
                fp_wkt = rt.get_footprint_wkt(ds, **rt_prm)
                footprint = GEOSGeometry(fp_wkt, srid)

                if isinstance(footprint, Polygon):
                    footprint = MultiPolygon(footprint)
                elif not isinstance(footprint, MultiPolygon):
                    raise TypeError(
                        "Got invalid geometry %s" % type(footprint).__name__
                    )

                values['footprint'] = footprint
            pass

        # --= dataset with no geocoding =--
        # TODO: Handling of other types of GDAL geocoding (e.g, RPC).
        else:
            pass

        reader = self._find_additional_reader(ds)
        if reader:
            additional_values = reader.read_ds(ds)
            for key, value in additional_values.items():
                values.setdefault(key, value)

        driver_metadata = driver.GetMetadata()
        frmt = driver_metadata.get("DMD_MIMETYPE")
        if frmt:
            values["format"] = frmt

        return values

    def _find_additional_reader(self, ds):
        for reader in get_gdal_dataset_format_readers():
            if reader.test_ds(ds):
                return reader
        return None



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import re
import functools
import json

from django.core.exceptions import FieldDoesNotExist

from eoxserver.core.decoders import kvp, enum
from eoxserver.core.util.xmltools import NameSpace
from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.services import filters
from eoxserver.resources.coverages import models


class EarthObservationExtension(object):
    """ Implementation of the OpenSearch `'EO' extension
    <http://docs.opengeospatial.org/is/13-026r8/13-026r8.html>`_.
    """

    namespace = NameSpace(
        "http://a9.com/-/opensearch/extensions/eo/1.0/", "eo"
    )

    def filter(self, qs, parameters):
        mapping, mapping_choices = filters.get_field_mapping_for_model(qs.model)
        decoder = EarthObservationExtensionDecoder(parameters)

        query_filters = []
        for filter_name, db_accessor in mapping.items():
            value = getattr(decoder, filter_name, None)

            if value:
                attr = filters.attribute(filter_name, mapping)
                if isinstance(value, list):
                    query_filters.append(filters.contains(attr, value))
                elif isinstance(value, dict):
                    if 'min' in value:
                        query_filters.append(
                            filters.compare(attr, value['min'],
                                '>=' if value['min_inclusive'] else '>',
                                mapping_choices
                            )
                        )
                    if 'max' in value:
                        query_filters.append(
                            filters.compare(attr, value['max'],
                                '<=' if value['max_inclusive'] else '<',
                                mapping_choices
                            )
                        )
                else:
                    query_filters.append(
                        filters.compare(attr, value, '=', mapping_choices)
                    )

        if query_filters:
            qs = qs.filter(
                filters.combine(query_filters, 'AND')
            )

        return qs

    def get_schema(self, collection=None, model_class=None):
        mapping, mapping_choices = filters.get_field_mapping_for_model(
            model_class or models.Product, True
        )

        schema = []
        summary = None
        if collection:
            summary = self._load_product_summary(collection)

        for key, value in mapping.items():
            param = dict(
                name=key, type=key
            )

            if summary:
                param_summary = summary.get(key)

                # leave out all parameters not present in the summary
                if not self._is_param_summary_valid(param_summary):
                    continue

                # insert information from the parameter summary
                if isinstance(param_summary, list):
                    param['options'] = param_summary
                elif isinstance(param_summary, dict):
                    min_ = param_summary.get('min')
                    max_ = param_summary.get('max')
                    if min_ is not None:
                        param['min'] = min_
                    if max_ is not None:
                        param['max'] = max_

            # use the mapping choices to get a list of options, if possible
            if 'options' not in param and value in mapping_choices:
                param['options'] = list(mapping_choices[value].keys())

            schema.append(param)

        return schema

    def _load_product_summary(self, collection):
        try:
            summary = json.loads(
                collection.collection_metadata.product_metadata_summary
            )
            return {
                filters._to_camel_case(key): value
                for key, value in summary.items()
            }
        except models.CollectionMetadata.DoesNotExist:
            pass
        return None

    def _is_param_summary_valid(self, param_summary):
        if not param_summary:
            return False

        elif isinstance(param_summary, dict):
            return param_summary.get('min') or param_summary.get('max')

        return True

    # def get_schema(self, collection):
    #     return [
    #         dict(
    #             name=key, type=key,
    #             pattern=,
    #             options=[
    #                 key for key in mapping_choices[value].keys()
    #             ] if value in mapping_choices else ()
    #         )
    #         for key, value in mapping.items()
    #     ]

float_pattern = r'[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?'
int_pattern = r'[-+]\d+'
datetime_pattern = r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(([+-]\d{2}:\d{2})|Z)'

base_range_pattern = (
    r'(?P<simple>%(base)s)$|'
    '(\{(?P<list>%(base)s(?:,%(base)s)*)\})$|'
    '(?P<range>(?P<low>[\[\]]%(base)s),(?P<high>%(base)s[\[\]]))$|'
    '(?P<only_low>[\[\]]%(base)s)$|'
    '(?P<only_high>%(base)s[\[\]])$'
)

float_range_pattern = re.compile(base_range_pattern % {"base": float_pattern})
int_range_pattern = re.compile(base_range_pattern % {"base": int_pattern})
datetime_range_pattern = re.compile(
    base_range_pattern % {"base": datetime_pattern}
)

def parse_range(value, pattern, value_parser):
    match = pattern.match(value)

    if match:
        values = match.groupdict()

        if values['simple']:
            return value_parser(values['simple'])
        elif values['list']:
            return [value_parser(v) for v in values['list'].split(',')]

        elif values['range']:
            low = values['low']
            hi = values['high']
            return {
                'min': value_parser(low[1:]),
                'min_inclusive': low[0] == "[",
                'max': value_parser(hi[:-1]),
                'max_inclusive': hi[-1] == "]"
            }
        elif values['only_low']:
            low = values['only_low']
            return {
                'min': value_parser(low[1:]),
                'min_inclusive': low[0] == "["
            }
        elif values['only_high']:
            hi = values['only_high']
            return {
                'max': value_parser(hi[:-1]),
                'max_inclusive': hi[-1] == "]"
            }

    return None


parse_float_range = functools.partial(
    parse_range, pattern=re.compile(float_range_pattern), value_parser=float
)

parse_int_range = functools.partial(
    parse_range, pattern=re.compile(float_range_pattern), value_parser=int
)

parse_datetime_range = functools.partial(
    parse_range, pattern=re.compile(datetime_range_pattern),
    value_parser=parse_iso8601
)


class EarthObservationExtensionDecoder(kvp.Decoder):
    productType = kvp.Parameter(num="?", type=str)
    doi = kvp.Parameter(num="?", type=str)
    platform = kvp.Parameter(num="?", type=str)
    platformSerialIdentifier = kvp.Parameter(num="?", type=str)
    instrument = kvp.Parameter(num="?", type=str)
    sensorType = kvp.Parameter(num="?", type=enum(('OPTICAL', 'RADAR', 'ALTIMETRIC', 'ATMOSPHERIC', 'LIMB'), False))
    compositeType = kvp.Parameter(num="?", type=str)
    processingLevel = kvp.Parameter(num="?", type=str)
    orbitType = kvp.Parameter(num="?", type=str)
    spectralRange = kvp.Parameter(num="?", type=str)
    wavelength = kvp.Parameter(num="?", type=parse_float_range)
    hasSecurityConstraints = kvp.Parameter(num="?", type=enum(('TRUE', 'FALSE'), False))
    dissemination = kvp.Parameter(num="?", type=str)
    recordSchema = kvp.Parameter(num="?", type=str)

    parentIdentifier = kvp.Parameter(num="?", type=str)
    productionStatus = kvp.Parameter(num="?", type=str)
    acquisitionType = kvp.Parameter(num="?", type=enum(('NOMINAL', 'CALIBRATION', 'OTHER'), False))
    orbitNumber = kvp.Parameter(num="?", type=parse_int_range)
    orbitDirection = kvp.Parameter(num="?", type=enum(('ASCENDING', 'DESCENDING'), False))
    track = kvp.Parameter(num="?", type=str)
    frame = kvp.Parameter(num="?", type=str)
    swathIdentifier = kvp.Parameter(num="?", type=str)
    cloudCover = kvp.Parameter(num="?", type=parse_int_range)
    snowCover = kvp.Parameter(num="?", type=parse_int_range)
    lowestLocation = kvp.Parameter(num="?", type=parse_float_range)
    highestLocation = kvp.Parameter(num="?", type=parse_float_range)
    productVersion = kvp.Parameter(num="?", type=str)
    productQualityStatus = kvp.Parameter(num="?", type=enum(('NOMINAL', 'DEGRADED'), False))
    productQualityDegradationTag = kvp.Parameter(num="?", type=str)
    processorName = kvp.Parameter(num="?", type=str)
    processingCenter = kvp.Parameter(num="?", type=str)
    creationDate = kvp.Parameter(num="?", type=parse_datetime_range)
    modificationDate = kvp.Parameter(num="?", type=parse_datetime_range)
    processingDate = kvp.Parameter(num="?", type=parse_datetime_range)
    sensorMode = kvp.Parameter(num="?", type=str)
    archivingCenter = kvp.Parameter(num="?", type=str)
    processingMode = kvp.Parameter(num="?", type=str)

    availabilityTime = kvp.Parameter(num="?", type=parse_datetime_range)
    acquisitionStation = kvp.Parameter(num="?", type=str)
    acquisitionSubType = kvp.Parameter(num="?", type=str)
    startTimeFromAscendingNode = kvp.Parameter(num="?", type=parse_int_range)
    completionTimeFromAscendingNode = kvp.Parameter(num="?", type=parse_int_range)
    illuminationAzimuthAngle = kvp.Parameter(num="?", type=parse_float_range)
    illuminationZenithAngle = kvp.Parameter(num="?", type=parse_float_range)
    illuminationElevationAngle = kvp.Parameter(num="?", type=parse_float_range)
    polarisationMode = kvp.Parameter(num="?", type=enum(('S', 'D', 'T', 'Q', 'UNDEFINED'), False))
    polarizationChannels = kvp.Parameter(num="?", type=str)
    antennaLookDirection = kvp.Parameter(num="?", type=str)
    minimumIncidenceAngle = kvp.Parameter(num="?", type=parse_float_range)
    maximumIncidenceAngle = kvp.Parameter(num="?", type=parse_float_range)
    dopplerFrequency = kvp.Parameter(num="?", type=parse_float_range)
    incidenceAngleVariation = kvp.Parameter(num="?", type=parse_float_range)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import os
from os.path import basename, join, relpath, split
from itertools import chain

from django.http.response import StreamingHttpResponse, FileResponse
import zipstream

from eoxserver.backends.access import vsi_open
from eoxserver.backends.storages import get_handler_for_model
from eoxserver.core.decoders import kvp, typelist, lower
from eoxserver.resources.coverages import models
from eoxserver.services.ows.dseo.v10.encoders import DSEO10CapabilitiesXMLEncoder


class MissingProductError(Exception):
    pass


class GetCapabilitiesHandler(object):
    service = 'DSEO'
    request = 'GetCapabilities'
    versions = ['1.0', '1.0.0']
    methods = ['GET']

    def handle(self, request):
        decoder = GetCapabilitiesKVPDecoder(request.GET)
        encoder = DSEO10CapabilitiesXMLEncoder()

        return encoder.serialize(
            encoder.encode_capabilities(request, decoder.sections),
            pretty_print=True
        ), encoder.content_type


class GetProductHandler(object):
    service = 'DSEO'
    request = 'GetProduct'
    versions = ['1.0', '1.0.0']
    methods = ['GET']

    def handle(self, request):
        decoder = GetProductKVPDecoder(request.GET)
        product_uri = decoder.product_uri

        try:
            product = models.Product.objects.get(identifier=product_uri)
        except models.Product.DoesNotExist:
            raise MissingProductError("Requested product is missing")

        package = product.package
        if package and package.parent is None:
            handler = get_handler_for_model(package)
            if handler.name in ('ZIP', 'TAR'):
                response = FileResponse(
                    open(package.url), content_type='application/octet-stream',
                )
                response['Content-Disposition'] = 'attachment; filename="%s"' % (
                    basename(package.url)
                )
                return response

            elif handler.name == 'directory':
                zip_stream = zipstream.ZipFile(
                    mode='w', compression=zipstream.ZIP_DEFLATED
                )
                # compute a base path name, in order to have the last part of
                # the path always in the filename
                base = split(
                    package.url[:-1] if package.url.endswith('/')
                    else package.url
                )[0]
                for root, _, filenames in os.walk(package.url):
                    for filename in filenames:
                        path = join(root, filename)
                        zip_stream.write(path, relpath(path, base))
                response = StreamingHttpResponse(
                    zip_stream, content_type='application/octet-stream'
                )
                response['Content-Disposition'] = \
                    'attachment; filename="%s.zip"' % product.identifier

                return response

        elif package:
            # TODO: determine whether the files are local. if yes then unpack
            # from parent storage
            raise NotImplementedError

        else:
            # for each coverage iterate over all metadata and array
            # metadata files and create a ZIP on the fly

            zip_stream = zipstream.ZipFile(
                mode='w', compression=zipstream.ZIP_DEFLATED
            )

            for coverage in product.coverages.all():
                items = chain(
                    coverage.arraydata_items.all(),
                    coverage.metadata_items.all()
                )

                for data_item in items:
                    # faster encoding when local files. Fallback on VSI API otherwise
                    if not data_item.storage:
                        zip_stream.write(
                            data_item.location,
                            join(
                                product.identifier, coverage.identifier,
                                basename(data_item.location)
                            )
                        )
                    else:
                        zip_stream.write_iter(
                            join(
                                product.identifier, coverage.identifier,
                                basename(data_item.location)
                            ),
                            _iter_data_item(data_item)
                        )

            response = StreamingHttpResponse(
                zip_stream, content_type='application/octet-stream'
            )
            response['Content-Disposition'] = 'attachment; filename="%s.zip"' % (
                product.identifier
            )
            return response


# helper function to iterate over a data item in chunks
def _iter_data_item(data_item, chunk_size=65535):
    with vsi_open(data_item) as vsi_file:
        while True:
            data = vsi_file.read(chunk_size)
            if data:
                yield data
            else:
                break


class GetCapabilitiesKVPDecoder(kvp.Decoder):
    sections = kvp.Parameter(type=typelist(lower, ","), num="?", default=["all"])
    updatesequence = kvp.Parameter(num="?")
    acceptversions = kvp.Parameter(type=typelist(str, ","), num="?")
    acceptformats = kvp.Parameter(type=typelist(str, ","), num="?", default=["text/xml"])
    acceptlanguages = kvp.Parameter(type=typelist(str, ","), num="?")


class GetProductKVPDecoder(kvp.Decoder):
    product_uri = kvp.Parameter('producturi', num=1)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import os.path
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
try:
    from ConfigParser import RawConfigParser
except ImportError:
    from configparser import RawConfigParser

from eoxserver.core.decoders import config
from django.utils.six import string_types

class NativeConfigFormatReader(object):
    def open_reader(self, obj):
        if isinstance(obj, string_types):
            try:
                parser = RawConfigParser()
                if os.path.exists(obj):
                    parser.read((obj,))
                else:
                    parser.readfp(StringIO(obj))

                return NativeConfigReader(parser)
            except:
                pass
        return None

    def test(self, obj):
        try:
            reader = self.open_reader(obj)
            reader.range_type_name
            return True
        except:
            return False

    def get_format_name(self, obj):
        return "native_config"

    def read(self, obj):
        reader = self.open_reader(obj)
        if reader:
            return {
                "range_type_name": reader.range_type_name
            }


class NativeConfigReader(config.Reader):
    range_type_name = config.Option(section="range_type")
